@@ -1,9 +1,10 @@
 
-__version__ = "7.20"
+__version__ = "8.20"
 __author__ = "Alex Al-Saffar"
-__date__ = "2022-08-01"
+__date__ = "2022-11-01"
 docs = r"https://crocodile.readthedocs.io/en/latest/"
 git = r"https://github.com/thisismygitrepo/crocodile"
+pypi = r"https://pypi.org/project/crocodile"
 release_notes = """
-Fixed bugs in SSH
+Minor fixes
 """
