@@ -4,22 +4,38 @@
 """
 
 import argparse
+import os
+
 from crocodile.core import *
 from crocodile.file_management import *
 from crocodile.meta import *
 # import crocodile.environment as env
 from crocodile.matplotlib_management import *
 import crocodile.toolbox as tb
-D = Display; L = List; E = Experimental; S = Struct
-
 import numpy as np
 import pandas as pd
+import platform
+import random
 
 
-print(f"Crocodile Shell {__import__('crocodile').__version__}")
+D = Display; L = List; E = Experimental; S = Struct
+
+
+print(f"Crocodile Shell {__import__('crocodile').__version__}. Python {platform.python_version()} On {platform.system()}.")
 # link to tutorial or github
 print(f"Made with ❤️")
-print(P(__file__).parent.joinpath("art").search().sample(size=1)[0].read_text())
+
+__ = P(__file__).parent.joinpath("art").search().sample(size=1)[0]
+if platform.system() == "Windows": print(__.read_text())
+else:
+    try:
+        surprise = random.choice([True, True, True, True, False])  # classic art (True) or boxes (False)
+        if surprise:
+            from crocodile.msc.ascii_art import get_art
+            artlib = random.choice(['boxes', 'cowsay'])
+            get_art("crocodile", calliagraphy=None, artlib=artlib, file=(__ := P.tmpfile("croco_art", folder="tmp_arts")), verbose=False)
+        os.system(f"cat {__} | lolcat")
+    except: print(__.read_text())
 
 
 def build_parser():
