@@ -4,6 +4,7 @@ from crocodile.matplotlib_management import ImShow, FigureSave
 import numpy as np
 import pandas as pd
 from abc import ABC
+from typing import Generic, TypeVar
 import enum
 from tqdm import tqdm
 import copy
@@ -11,6 +12,10 @@ import copy
 
 # %% ========================== DeepLearning Accessories =================================
 
+BM = TypeVar("BM", bound="BaseModel")
+DR = TypeVar("DR", bound="DataReader")
+HPM = TypeVar("HPM", bound="HyperParam")
+
 
 class Device(enum.Enum):
     gpu0 = 'gpu0'
@@ -55,7 +60,7 @@ class HyperParam(tb.Struct):
     def save(self, **kwargs):
         self.save_dir.joinpath(self.subpath / 'hparams.txt').create(parents_only=True).write_text(str(self))
         if self.save_type in {"data", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.dat.pkl"), add_suffix=False, data_only=True, desc="")
-        if self.save_type in {"obj", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.pkl"), add_suffix=False, data_only=False, desc="")
+        # if self.save_type in {"obj", "both"}: super(HyperParam, self).save(path=self.save_dir.joinpath(self.subpath / "hparams.HyperParam.pkl"), add_suffix=False, data_only=False, desc="")
 
     @classmethod
     def from_saved_data(cls, path, *args, **kwargs): return super(HyperParam, cls).from_saved_data(tb.P(path) / cls.subpath / "hparams.HyperParam.dat.pkl", *args, **kwargs)
@@ -148,54 +153,136 @@ class DataReader(tb.Base):
     implemented a fallback `getattr` method that allows accessing those attributes from the class data_only, without the 
     need to reference `.dataspects`.
     """
-    def __init__(self, hp: HyperParam = None, specs=None, split=None, *args, **kwargs):
+    def __init__(self, hp: Generic[HPM] = None, specs=None, split=None, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.hp = hp
         self.split = split
         self.plotter = None
         # attributes to be saved.
         self.specs = specs if specs else tb.Struct()
+        self.ip_strings = None  # e.g.: ["x1", "x2"]
+        self.op_strings = None  # e.g.: ["y1", "y2"]
+        self.other_strings = None  # e.g.: indices or names
+
+        # dataframes
         self.scaler = None
+        self.imputer = None
+        self.cols_ordinal = None
+        self.cols_onehot = None
+        self.cols_numerical = None
+        self.encoder_onehot = None
+        self.encoder_ordinal = None
 
     def save(self, path=None, *args, **kwargs):
         base = (tb.P(path) if path is not None else self.hp.save_dir).joinpath(self.subpath).create()
         if self.hp.save_type in {"data", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.dat.pkl", add_suffix=False, data_only=True)
-        if self.hp.save_type in {"obj", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.pkl", add_suffix=False, data_only=False)
+        # if self.hp.save_type in {"obj", "both"}: super(DataReader, self).save(path=base / "data_reader.DataReader.pkl", add_suffix=False, data_only=False)
 
     @classmethod
     def from_saved_data(cls, path, *args, **kwargs): return super(DataReader, cls).from_saved_data(tb.P(path) / cls.subpath / "data_reader.DataReader.dat.pkl", *args, **kwargs)
-    def __getstate__(self): return dict(specs=self.specs, scaler=self.scaler)
+    def __getstate__(self):
+        items = ["specs", "scaler", "imputer", "ip_strings", "op_strings", "other_strings", "cols_numerical", "cols_ordinal", "cols_onehot", "encoder_onehot", "encoder_ordinal"]
+        return dict(zip(items, [getattr(self, item) for item in items]))
     def __setstate__(self, state): return self.__dict__.update(state)
     def __repr__(self): return f"DataReader Object with these keys: \n" + tb.Struct(self.__dict__).print(as_config=False, return_str=True)
 
-    def split_the_data(self, *args, strings=None, **kwargs):
+    def split_the_data(self, *args, ip_strings=None, op_strings=None, others_string=None, **kwargs):
         from sklearn.model_selection import train_test_split
         result = train_test_split(*args, test_size=self.hp.test_split, shuffle=self.hp.shuffle, random_state=self.hp.seed, **kwargs)
         self.split = tb.Struct(train_loader=None, test_loader=None)
-        if strings is None: strings = ["x", "y"]
+        if ip_strings is None:
+            ip_strings = [f"x_{i}" for i in range(len(args)-1)]
+            if len(ip_strings) == 1: ip_strings = ["x"]
+        self.ip_strings = ip_strings
+        if op_strings is None: op_strings = ["y"]
+        self.op_strings = op_strings
+        if others_string is None: others_string = []
+        self.other_strings = others_string
+        strings = ip_strings + op_strings + others_string
+        self.specs.ip_shapes = []  # useful info for instantiating models.
+        self.specs.op_shapes = []
+        self.specs.other_shapes = []
+
+        for an_arg, key in zip(args, strings):
+            a_shape = an_arg.iloc[0].shape if type(an_arg) in {pd.DataFrame, pd.Series} else np.array(an_arg[0]).shape
+            if key in ip_strings: self.specs.ip_shapes.append(a_shape)
+            elif key in op_strings: self.specs.op_shapes.append(a_shape)
+            elif key in others_string: self.specs.other_shapes.append(a_shape)
         self.split.update({astring + '_train': result[ii * 2] for ii, astring in enumerate(strings)})
         self.split.update({astring + '_test': result[ii * 2 + 1] for ii, astring in enumerate(strings)})
-        x = self.split.x_train
-        self.specs.ip_shape = x.iloc[0].shape if type(x) is pd.DataFrame else x[0].shape  # useful info for instantiating models.
-        self.specs.op_shape = self.split.y_train[0].shape  # useful info for instantiating models.
         print(f"================== Training Data Split ===========================")
         self.split.print()
 
-    def sample_dataset(self, aslice=None, indices=None, use_default_slice=False, dataset="test"):
-        keys = self.split.keys().filter(lambda x: f'_{dataset}' in x)
-        selection = indices or aslice
-        res1 = selection or np.random.choice(len(self.split[keys[0]]), size=self.hp.batch_size, replace=False)
-        res2 = selection or slice(0, self.hp.batch_size)
-        selection = res1 if use_default_slice is False and (indices or aslice) is None else (res1 if indices is not None else res2)
-        return tuple([tmp.iloc[selection] if type(tmp := self.split[key]) is pd.DataFrame else tmp[selection] for key in keys])
+    def get_data_strings(self, which_data="ip", which_split="train"):
+        strings = {"op": self.op_strings, "ip": self.ip_strings, "others": self.other_strings}[which_data]
+        keys_ip = [item + f"_{which_split}" for item in strings]
+        return keys_ip
 
-    def get_random_input_output(self, ip_shape=None, op_shape=None):
-        if ip_shape is None: ip_shape = self.specs.ip_shape
-        if op_shape is None: op_shape = self.specs.op_shape
+    def sample_dataset(self, aslice=None, indices=None, use_slice=False, split="test", size=None):
+        assert self.split is not None, f"No dataset is loaded to DataReader, .split attribute is empty. Consider using `.load_training_data()` method."
+        keys_ip = self.get_data_strings(which_data="ip", which_split=split)
+        keys_op = self.get_data_strings(which_data="op", which_split=split)
+        keys_others = self.get_data_strings(which_data="others", which_split=split)
+        ds_size = len(self.split[keys_ip[0]])
+        select_size = size or self.hp.batch_size
+        start_idx = np.random.choice(ds_size - select_size)
+
+        if indices is not None: selection = indices
+        elif aslice is not None: selection = aslice
+        elif use_slice: selection = slice(start_idx, start_idx + select_size)  # ragged tensors don't support indexing, this can be handy in that case.
+        else: selection = np.random.choice(ds_size, size=select_size, replace=False)
+
+        x, y, others = [], [], []
+        for idx, key in zip([0] * len(keys_ip) + [1] * len(keys_op) + [2] * len(keys_others), keys_ip + keys_op + keys_others):
+            tmp = self.split[key]
+            item = tmp.iloc[selection] if type(tmp) in {pd.DataFrame, pd.Series} else tmp[selection]
+            if idx == 0: x.append(item)
+            elif idx == 1: y.append(item)
+            else: others.append(item)
+        x = x[0] if len(self.ip_strings) == 1 else x
+        y = y[0] if len(self.op_strings) == 1 else y
+        others = others[0] if len(self.other_strings) == 1 else others
+        if len(others) == 0:
+            # others = np.arange(len(x if len(self.ip_strings) == 1 else x[0]))
+            if type(selection) is slice:
+                others = np.arange(*selection.indices(10000000000000))
+            else:
+                others = selection
+        return x, y, others
+
+    def get_random_inputs_outputs(self, ip_shapes=None, op_shapes=None):
+        if ip_shapes is None: ip_shapes = self.specs.ip_shapes
+        if op_shapes is None: op_shapes = self.specs.op_shapes
         dtype = self.hp.precision if hasattr(self.hp, "precision") else "float32"
-        ip = np.random.randn(*((self.hp.batch_size,) + ip_shape)).astype(dtype)
-        op = np.random.randn(*((self.hp.batch_size,) + op_shape)).astype(dtype)
-        return ip, op
+        x = [np.random.randn(*((self.hp.batch_size,) + ip_shape)).astype(dtype) for ip_shape in ip_shapes]
+        y = [np.random.randn(*((self.hp.batch_size,) + op_shape)).astype(dtype) for op_shape in op_shapes]
+        x = x[0] if len(self.ip_strings) == 1 else x
+        y = y[0] if len(self.op_strings) == 1 else y
+        return x, y
+
+    def profile_dataframe(self, df, file=None, silent=False, suffix="", explorative=True):
+        profile_report = tb.install_n_import("pandas_profiling").ProfileReport
+        # from import ProfileReport  # also try pandasgui  # import statement is kept inside the function due to collission with matplotlib
+        file = file or self.hp.save_dir.joinpath(self.subpath, f"pandas_profile_report_{suffix}.html").create(parents_only=True)
+        profile_report(df, title="Pandas Profiling Report", explorative=explorative).to_file(file, silent=silent)
+        return file
+    def open_dataframe_profile(self): self.hp.save_dir.joinpath(self.subpath, "pandas_profile_report.html")()
+
+    def encode(self, df: pd.DataFrame) -> pd.DataFrame:
+        """Converts the dataframe to numerical format. Missing values are encoded as `pd.NA`, otherwise, encoders will fail to handle them."""
+        df[self.cols_ordinal] = self.encoder_ordinal.transform(df[self.cols_ordinal])
+        tmp = self.encoder_onehot.transform(df[self.cols_onehot])
+        df.drop(columns=self.cols_onehot, inplace=True)
+        df[self.encoder_onehot.get_feature_names_out()] = tmp
+        df[self.cols_numerical] = df[self.cols_numerical].to_numpy().astype(self.hp.precision)
+        return df
+
+    def impute_standardize(self, df: pd.DataFrame) -> pd.DataFrame:
+        df.fillna(np.nan, inplace=True)  # SKlearn Imputer only works with Numpy's np.nan, as opposed to Pandas' pd.NA
+        columns = df.columns
+        df = self.imputer.transform(df)
+        df = self.scaler.transform(pd.DataFrame(df, columns=columns))
+        return pd.DataFrame(df, columns=columns)
 
     def preprocess(self, *args, **kwargs): _ = args, kwargs, self; return args[0]  # acts like identity.
     def postprocess(self, *args, **kwargs): _ = args, kwargs, self; return args[0]  # acts like identity
@@ -233,13 +320,14 @@ class BaseModel(ABC):
     Functionally or Sequentually built models are much more powerful than Subclassed models. They are faster, have more features, can be plotted, serialized, correspond to computational graphs etc.
     """
     # @abstractmethod
-    def __init__(self, hp=None, data=None, model=None, compiler=None, history=None):
+    def __init__(self, hp: Generic[HPM] = None, data: Generic[BM] = None, model=None, compiler=None, history=None):
         self.hp = hp  # should be populated upon instantiation.
         self.model = model  # should be populated upon instantiation.
         self.data = data  # should be populated upon instantiation.
         self.compiler = compiler  # Struct with .losses, .metrics and .optimizer.
         self.history = tb.List() if history is None else history  # should be populated in fit method, or loaded up.
         self.plotter = FigureSave.NullAuto
+        self.fig = None
         self.kwargs = None
         self.tmp = None
 
@@ -266,14 +354,23 @@ class BaseModel(ABC):
         # in both cases: pass the specs to the compiler if we have TF framework
         if self.hp.pkg.__name__ == "tensorflow" and compile_model: self.model.compile(**self.compiler.__dict__)
 
-    def fit(self, viz=False, **kwargs):
-        default_settings = tb.Struct(x=self.data.split.x_train, y=self.data.split.y_train, validation_data=(self.data.split.x_test, self.data.split.y_test),
+    def fit(self, viz=True, val_sample_weights=None, **kwargs):
+        x_train = self.data.split.get(keys=self.data.get_data_strings(which_data="ip", which_split="train")).list
+        y_train = self.data.split.get(keys=self.data.get_data_strings(which_data="op", which_split="train")).list
+        x_test = self.data.split.get(keys=self.data.get_data_strings(which_data="ip", which_split="test")).list
+        y_test = self.data.split.get(keys=self.data.get_data_strings(which_data="op", which_split="test")).list
+        x_test = x_test[0] if len(x_test) == 1 else x_test
+        y_test = y_test[0] if len(y_test) == 1 else y_test
+        default_settings = tb.Struct(x=x_train[0] if len(x_train) == 1 else x_train,
+                                     y=y_train[0] if len(y_train) == 1 else y_train,
+                                     validation_data=(x_test, y_test) if val_sample_weights is None else (x_test, y_test, val_sample_weights),
                                      batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1, shuffle=self.hp.shuffle, callbacks=[])
         default_settings.update(kwargs)
         hist = self.model.fit(**default_settings.dict)
-        self.history.append(tb.Struct(copy.deepcopy(hist.history)))
-        # it is paramount to copy, cause source can change.
-        if viz: self.plot_loss()
+        self.history.append(tb.Struct(copy.deepcopy(hist.history)))  # it is paramount to copy, cause source can change.
+        if viz:
+            artist = self.plot_loss()
+            artist.fig.savefig(self.hp.save_dir.joinpath(f"metadata/loss_curve.png").append(index=True).create(parents_only=True))
         return self
 
     def switch_to_sgd(self, epochs=10):
@@ -301,15 +398,23 @@ class BaseModel(ABC):
 
     def postprocess(self, *args, **kwargs): return self.data.postprocess(*args, **kwargs)
     def __call__(self, *args, **kwargs): return self.model(*args, **kwargs)
-    def viz(self, *args, **kwargs): self.data.viz(*args, **kwargs)
+    def viz(self, *args, **kwargs): return self.data.viz(*args, **kwargs)
     def save_model(self, directory): self.model.save(directory)  # In TF: send only path dir. Save path is saved_model.pb
     def save_weights(self, directory): self.model.save_weights(directory.joinpath(self.model.name))  # TF: last part of path is file path.
     @staticmethod
     def load_model(directory): __import__("tensorflow").keras.models.load_model(directory)  # path to directory. file saved_model.pb is read auto.
     def load_weights(self, directory): self.model.load_weights(directory.glob('*.data*').__next__().__str__().split('.data')[0])  # requires path to file path.
-    def summary(self): return self.model.summary()
+    def summary(self):
+        from contextlib import redirect_stdout
+        path = self.hp.save_dir.joinpath("metadata/model_summary.txt").create(parents_only=True)
+        with open(str(path), 'w') as f:
+            with redirect_stdout(f): self.model.summary()
+        return self.model.summary()
     def config(self): [print(layer.get_config(), "\n==============================") for layer in self.model.layers]; return None
-    def plot_loss(self, *args, **kwargs): return tb.Struct.concat_values(*self.history).plot(*args, **kwargs)
+    def plot_loss(self, *args, **kwargs): return tb.Struct.concat_values(*self.history).plot(*args, title="Loss Curve",
+                                                                                             xlabel="epochs",
+                                                                                             ylabel=self.compiler.loss.name if hasattr(self.compiler.loss, "name") else self.compiler.loss.__name__,
+                                                                                             **kwargs)
 
     def infer(self, x):
         """ This method assumes numpy input, datatype-wise and is also preprocessed.
@@ -319,6 +424,10 @@ class BaseModel(ABC):
         """
         return self.model.predict(x)  # Keras automatically handles special layers, can accept dataframes, and always returns numpy.
 
+    def predict(self, x, **kwargs):
+        """This method assumes preprocessed input. Returns postprocessed output. It is useful at evaluation time with preprocessed test set."""
+        return self.postprocess(self.infer(x), **kwargs)
+
     def deduce(self, obj, viz=True, **kwargs):
         """Assumes that contents of the object are in the form of a batch."""
         preprocessed = self.preprocess(obj, **kwargs)
@@ -328,45 +437,42 @@ class BaseModel(ABC):
         if viz: self.viz(postprocessed, **kwargs)
         return result
 
-    def predict(self, x, **kwargs):
-        """This method assumes preprocessed input. Returns postprocessed output. It is useful at evaluation time with preprocessed test set."""
-        return self.postprocess(self.infer(x), **kwargs)
-
-    def evaluate(self, x_test=None, y_test=None, names_test=None, idx=None, viz=True, sample=5, **kwargs):
-        # ================= Data Procurement ===================================
-        x_test, y_test = x_test if x_test is not None else self.data.split.x_test, y_test if y_test is not None else self.data.split.y_test
-        names_test = names_test if names_test is not None else (self.data.split.names_test if hasattr(self.data.split, "names_test") else np.arange(len(x_test)))
-        idx = np.random.choice(len(x_test) - 1, size=sample, replace=False) if idx is None else (slice(idx, idx + 1, 1) if isinstance(idx, int) else idx)
-        x_test, y_test, names_test = x_test[idx] if "DataFrame" not in str(type(x_test)) else x_test.iloc[idx], y_test[idx] , names_test[idx]
+    def evaluate(self, x_test=None, y_test=None, names_test=None, aslice=None, indices=None, use_slice=False, size=None, split="test", viz=True, **kwargs):
+        if x_test is None and y_test is None and names_test is None:
+            x_test, y_test, names_test = self.data.sample_dataset(aslice=aslice, indices=indices, use_slice=use_slice, split=split, size=size)
+        elif names_test is None: names_test = np.arange(len(x_test))
         # ==========================================================================
-        prediction = self.infer(x_test)
-        loss_dict = self.get_metrics_evaluations(prediction, y_test)
-        if loss_dict is not None: loss_dict['names'] = names_test
-        pred = self.postprocess(prediction, per_instance_kwargs=dict(name=names_test), legend="Prediction", **kwargs)
-        gt = self.postprocess(y_test, per_instance_kwargs=dict(name=names_test), legend="Ground Truth", **kwargs)
-        results = tb.Struct(pp_prediction=pred, prediction=prediction, input=x_test, pp_gt=gt, gt=y_test, names=names_test, loss_df=loss_dict, )
+        y_pred = self.infer(x_test)
+        loss_df = self.get_metrics_evaluations(y_pred, y_test)
+        if loss_df is not None:
+            if len(self.data.other_strings) == 1: loss_df[self.data.other_strings[0]] = names_test
+            else:
+                for val, name in zip(names_test, self.data.other_strings): loss_df[name] = val
+        y_pred_pp = self.postprocess(y_pred, per_instance_kwargs=dict(name=names_test), legend="Prediction", **kwargs)
+        y_true_pp = self.postprocess(y_test, per_instance_kwargs=dict(name=names_test), legend="Ground Truth", **kwargs)
+        results = tb.Struct(x=x_test, y_pred=y_pred, y_pred_pp=y_pred_pp, y_true=y_test, y_true_pp=y_true_pp, names=names_test, loss_df=loss_df, )
         if viz:
             loss_name = results.loss_df.columns.to_list()[0]  # first loss path
             loss_label = results.loss_df[loss_name].apply(lambda x: f"{loss_name} = {x}").to_list()
             names = [f"{aname}. Case: {anindex}" for aname, anindex in zip(loss_label, names_test)]
-            self.viz(pred, gt, names=names, **kwargs)
+            self.fig = self.viz(y_pred_pp, y_true_pp, names=names, **kwargs)
         return results
 
-    def get_metrics_evaluations(self, prediction, groun_truth):
+    def get_metrics_evaluations(self, prediction, groun_truth) -> pd.DataFrame or None:
         if self.compiler is None: return None
         metrics = tb.L([self.compiler.loss]) + self.compiler.metrics
         loss_dict = dict()
         for a_metric in metrics:
-            if hasattr(a_metric, "path"): name = a_metric.name
+            if hasattr(a_metric, "name"): name = a_metric.name
             elif hasattr(a_metric, "__name__"): name = a_metric.__name__
-            else: name = "unknown"
+            else: name = "unknown_loss_name"
             # try:  # EAFP vs LBYL: both are duck-typing styles as they ask for what object can do (whether by introspection or trial) as opposed to checking its type.
             #     path = a_metric.path  # works for subclasses Metrics
             # except AttributeError: path = a_metric.__name__  # works for functions.
             loss_dict[name] = []
             for a_prediction, a_y_test in zip(prediction, groun_truth):
                 if hasattr(a_metric, "reset_states"): a_metric.reset_states()
-                loss = a_metric(a_prediction[None], a_y_test[None])
+                loss = a_metric(y_pred=a_prediction[None], y_true=a_y_test[None])
                 loss_dict[name].append(np.array(loss).item())
         return pd.DataFrame(loss_dict)
 
@@ -382,17 +488,19 @@ class BaseModel(ABC):
         self.hp.save()  # goes into the meta path.
         self.data.save()  # goes into the meta path.
         tb.Save.pickle(obj=self.history, path=self.hp.save_dir / 'metadata/history.pkl', verbose=True, desc="Training History")  # goes into the meta path.
-        tb.Experimental.generate_readme(self.hp.save_dir, obj=self.__class__, **kwargs)
+        try: tb.Experimental.generate_readme(self.hp.save_dir, obj=self.__class__, **kwargs)
+        except Exception as ex: print(ex)  # often fails because model is defined in main during experiments.
         save_dir = self.hp.save_dir.joinpath(f'{"weights" if weights_only else "model"}_save_v{version}').create()  # model save goes into data path.
         if weights_only: self.save_weights(save_dir)
         else: self.save_model(save_dir)
         print(f'SAVED Model Class @ {self.hp.save_dir.as_uri()}')
+        return self.hp.save_dir
 
     @classmethod
     def from_class_weights(cls, path, hparam_class=None, data_class=None, device_name=None, verbose=True):
         path = tb.P(path)
         if hparam_class is not None: hp_obj = hparam_class.from_saved_data(path)
-        else: hp_obj = (path / HyperParam.subpath + ".HyperParam.pkl").readit()
+        else: hp_obj = (path / HyperParam.subpath + "hparams.HyperParam.pkl").readit()
         if device_name: hp_obj.device_name = device_name
         if data_class is not None: d_obj = data_class.from_saved_data(path, hp=hp_obj)
         else: d_obj = (path / DataReader.subpath / "data_reader.DataReader.pkl").readit()
@@ -413,33 +521,55 @@ class BaseModel(ABC):
         wrapper_class = cls(hp_obj, data_obj, model_obj)
         return wrapper_class
 
-    def plot_model(self, **kwargs):  # alternative viz via tf2onnx then Netron.
+    def plot_model(self, dpi=150, **kwargs):  # alternative viz via tf2onnx then Netron.
         import tensorflow as tf
-        tf.keras.utils.plot_model(self.model, to_file=self.hp.save_dir / 'model_plot.png', show_shapes=True, show_layer_names=True, show_dtype=True, expand_nested=True, dpi=150, **kwargs)
-        print(f"Successfully plotted the model @ {(self.hp.save_dir / 'model_plot.png').as_uri()}")
+        path = self.hp.save_dir / 'model_plot.png'
+        tf.keras.utils.plot_model(self.model, to_file=path, show_shapes=True, show_layer_names=True, show_layer_activations=True, show_dtype=True, expand_nested=True, dpi=dpi, **kwargs)
+        print(f"Successfully plotted the model @ {path.as_uri()}")
+        return path
 
-    def build(self, ip_shape=None, verbose=True):
+    def build(self, sample_dataset=False, ip_shapes=None, ip=None, verbose=True):
         """ Building has two main uses.
         * Useful to baptize the model, especially when its layers are built lazily. Although this will eventually happen as the first batch goes in. This is a must before showing the summary of the model.
         * Doing sanity check about shapes when designing model.
         * Sanity check about values and ranges when random normal input is fed.
-        :param ip_shape:
+        :param sample_dataset:
+        :param ip_shapes:
+        :param ip:
         :param verbose:
         :return:
         """
-        ip, _ = self.data.get_random_input_output(ip_shape=ip_shape)
-        self.tmp = self.model(ip)  # op
+        keys_ip = self.data.get_data_strings(which_data="ip", which_split="test")
+        keys_op = self.data.get_data_strings(which_data="op", which_split="test")
+
+        if ip is None:
+            if sample_dataset: ip, _, _ = self.data.sample_dataset()
+            else: ip, _ = self.data.get_random_inputs_outputs(ip_shapes=ip_shapes)
+        op = self.model(inputs=ip)
+        ops = [op] if len(keys_op) == 1 else op
+        ips = [ip] if len(keys_ip) == 1 else ip
         if verbose:
+            print("\n")
             print("Build Test".center(50, '-'))
-            print(f"Input shape = {ip.shape}")
-            print(f"Output shape = {self.tmp.shape}")
-            print("Stats on output data for random normal input:")
-            print(pd.DataFrame(np.array(self.tmp).flatten()).describe())
-            print("-" * 100 + '\n' * 2)
+            print(f"Input shapes:")
+            tb.Struct.from_keys_values(keys_ip, tb.L(ips).apply(lambda x: x.shape)).print(as_config=True)
+            print(f"Output shape:")
+            tb.Struct.from_keys_values(keys_op, tb.L(ops).apply(lambda x: x.shape)).print(as_config=True)
+            print("\n\nStats on output data for random normal input:")
+            try:
+                res = []
+                for item_str, item_val in zip(keys_ip + keys_op, list(ips) + list(ops)):
+                    a_df = pd.DataFrame(np.array(item_val).flatten()).describe().rename(columns={0: item_str})
+                    res.append(a_df)
+                print(pd.concat(res, axis=1))
+            except Exception as ex:
+                print(f"Could not do stats on outputs and inputs. Error: {ex}")
+            print("Build Test Finished".center(50, '-'))
+            print("\n")
 
 
 class Ensemble(tb.Base):
-    def __init__(self, hp_class=None, data_class=None, model_class=None, size=15, *args, **kwargs):
+    def __init__(self, hp_class: Generic[HPM] = None, data_class: Generic[DR] = None, model_class: Generic[BM] = None, size=10, *args, **kwargs):
         """
         :param model_class: Either a class for constructing saved_models or list of saved_models already cosntructed.
           * In either case, the following methods should be implemented:
@@ -454,10 +584,11 @@ class Ensemble(tb.Base):
         self.hp_class = hp_class
         self.data_class = data_class
         self.model_class = model_class
+        self.models = tb.List()
+        self.data = None  # one data object for all models (so that it can fit in the memory)
         if hp_class and data_class and model_class:
-            self.models = tb.List()
             # only generate the dataset once and attach it to the ensemble to be reused by models.
-            self.data = self.data_class(hp_class())
+            self.data = self.data_class(hp_class(), load_trianing_data=True)
             print("Creating Models".center(100, "="))
             for i in tqdm(range(size)):
                 hp = self.hp_class()
@@ -483,12 +614,14 @@ class Ensemble(tb.Base):
         self.performance = tb.L()
         for i in range(self.size):
             print('\n\n', f" Training Model {i} ".center(100, "*"), '\n\n')
-            if shuffle_train_test: self.data.split_my_data_now(seed=np.random.randint(0, 1000))  # shuffle data (shared among models)
+            if shuffle_train_test:
+                self.models[i].hp.seed = np.random.randint(0, 1000)
+                self.data.split_the_data()  # shuffle data (shared among models)
             self.models[i].fit(**kwargs)
             self.performance.append(self.models[i].evaluate(idx=slice(0, -1), viz=False))
             if save:
                 self.models[i].save_class()
-                self.performance.save_pickle(self.hp_class.save_dir / "performance.pkl")
+                self.performance.save(self.hp_class.save_dir / "performance.List.pkl")
         print("\n\n", f" Finished fitting the ensemble ".center(100, ">"), "\n")
 
     def clear_memory(self): pass  # t.cuda.empty_cache()
@@ -628,5 +761,10 @@ def batcherv2(func_type='function', order=1):
         return Batch
 
 
+def get_template():
+    tb.install_n_import("clipboard").copy(tb.P(__file__).parent.joinpath("msc/dl_template.py").read_text(encoding="utf-8"))
+    print("Copied to clipboard")
+
+
 if __name__ == '__main__':
     pass
