@@ -8,9 +8,11 @@ import sys
 P = tb.P
 L = tb.L
 
-system = platform.system()
+system = platform.system()  # Linux or Windows
+myhome = tb.P(f"myhome/{platform.system().lower()}")
+
 OS = os.getenv("OS")  # Windows_NT
-sep = ";" if system == "Windows" else ":"  # PATH separator, this is special for PATH object, not to be confused with P.sep (normal paths).
+sep = ";" if system == "Windows" else ":"  # PATH separator, this is special for PATH object, not to be confused with P.sep (normal paths), usually / or \
 env = tb.Struct(dict(os.environ)).clean_view
 exe = P(sys.executable)
 
@@ -46,14 +48,19 @@ LogonServer       = os.getenv("LOGONSERVER")  # e.g. "\\MY-SURFACE"
 # HomePath          = P(tmp) if (tmp := os.getenv("HOMEPATH")) else None  # e.g. C:\Users\username
 Public            = P(tmp) if (tmp := os.getenv("PUBLIC")) else None  # C:\Users\Public
 
-WSL_FROM_WIN = P(r"\\wsl.localhost\Ubuntu-22.04")  # P(rf"\\wsl$\Ubuntu\home")  # see localappdata/canonical
+WSL_FROM_WIN = P(r"\\wsl.localhost\Ubuntu-22.04\home")  # P(rf"\\wsl$\Ubuntu\home")  # see localappdata/canonical
 WIN_FROM_WSL = P(rf"/mnt/c/Users")
 
 
 OneDriveConsumer = P(tmp) if (tmp := os.getenv("OneDriveConsumer")) else None
 OneDriveCommercial = P(tmp) if (tmp := os.getenv("OneDriveCommercial")) else None
 OneDrive = P(tmp) if (tmp := os.getenv("OneDrive")) else None
-OneDriveExe = LocalAppData.joinpath("Microsoft/OneDrive/OneDrive.exe") if LocalAppData else None
+if system == "Windows":
+    tmp1 = LocalAppData.joinpath("Microsoft/OneDrive/OneDrive.exe")
+    tmp2 = P(r"C:/Program Files/Microsoft OneDrive/OneDrive.exe")
+    OneDriveExe = tmp1 if tmp1.exists() else tmp2
+else: OneDriveExe = None
+
 
 DotFiles = P.home().joinpath("dotfiles")
 
@@ -126,18 +133,25 @@ class EnvVar:
 
 class PathVar:
     @staticmethod
-    def append_temporarily(path, kind="append", run=False):
+    def append_temporarily(dirs, kind="append", run=False):
+        dirs_ = []
+        for path in dirs:
+            path = tb.P(path).collapseuser(strict=False)
+            path = path.as_posix() if system == "Linux" else str(path)
+            if path in Path: print(f"Path passed `{path}` is already in PATH, skipping the appending.")
+            else: dirs_.append(path)
+        dirs = dirs_
+        if len(dirs) == 0: return ""
+
         if system == "Windows":
             """Source: https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_environment_variables?view=powershell-7.2"""
-            if tb.P(path) in Path:
-                print(f"Path passed `{path}` is already in PATH, skipping the appending.")
-                return None
-            if kind == "append": command = fr'$env:Path += ";{path}"'  # Append to the Path variable in the current window:
-            elif kind == "prefix": command = fr'$env:Path = "{path};" + $env:Path'  # Prefix the Path variable in the current window:
-            elif kind == "replace": command = fr'$env:Path = "{path}"'  # Replace the Path variable in the current window (use with caution!):
+            if kind == "append": command = fr'$env:Path += ";{sep.join(dirs)}"'  # Append to the Path variable in the current window:
+            elif kind == "prefix": command = fr'$env:Path = "{sep.join(dirs)};" + $env:Path'  # Prefix the Path variable in the current window:
+            elif kind == "replace": command = fr'$env:Path = "{sep.join(dirs)}"'  # Replace the Path variable in the current window (use with caution!):
             else: raise KeyError
             return command if run is False else tm.run(command, shell="powershell")
-        else: result = f'export PATH="{path}:$PATH"'
+        elif system == "Linux": result = f'export PATH="{sep.join(dirs)}:$PATH"'
+        else: raise ValueError
         return result if run is False else tm.run(result, shell="powershell")
 
     @staticmethod
@@ -190,7 +204,7 @@ def get_shell_profiles(shell):
 
 
 def construct_path(path_list): return tb.L(__import__("pd").unique(path_list)).reduce(lambda x, y: str(x) + sep + str(y))
-def get_path_defined_files(string_="*.exe"): res = tb.P.get_env().Path.search(string_).reduce(lambda x, y: x + y); res.print(); return res
+def get_path_defined_files(string_="*.exe"): res = Path.search(string_).reduce(lambda x, y: x + y); res.print(); return res
 
 
 if __name__ == '__main__':
